@@ -1504,8 +1504,7 @@ def simple_build(name: Union[TestName, str],
     # Required by GHC 7.3+, harmless for earlier versions:
     if (getTestOpts().c_src or
         getTestOpts().objc_src or
-        getTestOpts().objcpp_src or
-        getTestOpts().cmm_src):
+        getTestOpts().objcpp_src):
         extra_hc_opts += ' -no-hs-main '
 
     if getTestOpts().compile_cmd_prefix == '':
